// How to compile 
// mex rs232cj2.cpp  -f msvc60opts.bat
// put /MT option in msvc60opts.bat like as follows
//set OPTIMFLAGS=/MT -O2 -Oy- -DNDEBUG
//set DEBUGFLAGS=/MTd -Zi -Fd"%OUTDIR%%MEX_NAME%.pdb"
// Copyright(c) Gerox 1999 July 2 Kazuyuki Kobayashi
// support Windows 2000    April 13 Kazuyuki Kobayashi
// support flowcontrol 2007/May/3
// support upto COM99  2007/May/11
// Optimize code 2007/Aug/02
// Bug fix and increse BUFSIZE for Hokyo ME command 2010/Sep/06
// Optimize for speed 2010/Sep/10
// support Sleep option 2010/Sep/11

#include <windows.h>
#include <process.h>
#include "mex.h"
#define print mexPrintf("%s %d\n",__FILE__,__LINE__)
#pragma comment(lib,"user32.lib")
#define TIMEOUT_MUL 50
#define BUFSIZE 0x3fff

static DCB dcb;
static COMMTIMEOUTS timeouts;

static int initialized = 0;
static HANDLE hThrd = NULL;
static CRITICAL_SECTION Cs;
static OVERLAPPED _olRead;
static OVERLAPPED _olWrite;

//static unsigned char *rdata = NULL;
volatile static unsigned char rdata[BUFSIZE+1];
volatile static int rlen = 0;
static HANDLE hf_com;

//static DWORD dwThrdID;
static unsigned int dwThrdID;
static int flag = 0;
static int sleepinterval = 100;
static void showusage(void) {
	mexPrintf("-------------------------------------------------------\n");
	mexPrintf("rs232cj2(Port,BaudRate,ByteSize,Parity,StopBits,FlowCtl,Sleep)\n");
	mexPrintf("Example:   >>rs232cj2('COM1' ,9600,8,0,1,0);\n");
	mexPrintf("For   Gyro:>>rs232cj2('COM1' ,9600,8,1,1,0);\n");
	mexPrintf("For   KHR1:>>rs232cj2('COM1' ,9600,8,0,1,4);\n");
	mexPrintf("For Hokuyo:>>rs232cj2('COM1' ,19200,8,0,1,0,50);\n");
	mexPrintf("For   SICK:>>rs232cj2('COM1' ,500000,8,0,1,0);\n");
	mexPrintf("-------------------------------------------------------\n");
	mexPrintf("Port    : 'COM1','COM2',...,'COM10',... up to 'COM99'\n");
	mexPrintf("BaudRate: 600,1200,2400,4800,9600,14400,19200,38400,57600,115200\n");
	mexPrintf("ByteSize: 8,7\n");
	mexPrintf("Parity  : 0(none),1(odd),2(even),3(mark),4(space)\n");
	mexPrintf("StopBits: 1(1bit),1.5(1.5bit),2(2bit)\n");
	mexPrintf("FlowCtl : 0(none),1(Xon/Xoff),2(hardware),3(disable)\n");
	mexPrintf("        : 4(RTS=0,DTR=0),5(RTS=0,DTR=1),6(RTS=1,DTR=0),7(RTS=1,DTR=1)\n");
	mexPrintf("Sleep   : 0-32767[mS]\n");
	mexPrintf("            0(no sleep interval consumes high cpu power)\n");
	mexPrintf("          100(100mS sleep interval requires less cpu power)defalut\n");
	mexPrintf("Ring buffer size = %d bytes\n",BUFSIZE);
	mexPrintf("-------------------------------------------------------\n");
	mexPrintf("Init    : rs232cj2('COM1',BaudRate,ByteSize,Parity,StopBits,Flwctl,Sleep)\n");
	mexPrintf("Getc    : uint8_data = rs232cj2([]);\n");
	mexPrintf("Putc    : status = rs232cj2(uint8_data);\n");
	mexPrintf("Getflag : [CTS,DSR,RING,CD] = rs232cj2;\n");
	mexPrintf("-------------------------------------------------------\n");
	mexPrintf("RS232C asynchronous communication program by Gerox\n"); 
	mexPrintf("Windows 2000/XP/Vista/7 with MATLAB 7.1 or later \n");
	mexPrintf("Copyright 1999-2010 Gerox(c) Ikko. K. Kobayashi\n");
	mexPrintf("[%s] Build %s http://www.gerox.com\n",__FILE__,__DATE__);
	mexPrintf("-------------------------------------------------------\n");
}

static void errorfunc(int line) {
	mexPrintf("Error in %s[%d]\n",__FILE__,line);
	showusage();
	mexErrMsgTxt(" ");
}
static int initdcb(char* Port) {
	hf_com = CreateFile(Port, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_FLAG_WRITE_THROUGH |FILE_FLAG_NO_BUFFERING | FILE_FLAG_OVERLAPPED, NULL);
	if(hf_com == INVALID_HANDLE_VALUE){
		CHAR ErrorBuf[256];
		wsprintf(ErrorBuf,"CreateFile Failed GetLastError = %d",GetLastError());
		mexPrintf("%s %s line %d %s\n",ErrorBuf,__FILE__,__LINE__,__DATE__);
		return -1;
	}
	GetCommState(hf_com,&dcb);
	GetCommTimeouts(hf_com,&timeouts);
	CloseHandle(hf_com);
	DWORD DErr;
	ClearCommError(hf_com,&DErr,NULL);
	SetupComm(hf_com, BUFSIZE, BUFSIZE);
	PurgeComm(hf_com, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR ) ;
	dcb.fBinary = 1;
	dcb.fAbortOnError = false;
	dcb.EvtChar = 0x0d;
	timeouts.ReadIntervalTimeout = MAXDWORD ; 
	timeouts.ReadTotalTimeoutMultiplier = TIMEOUT_MUL;
	timeouts.ReadTotalTimeoutConstant = 0 ;
	timeouts.WriteTotalTimeoutMultiplier = TIMEOUT_MUL;
	timeouts.WriteTotalTimeoutConstant = 0 ;
	SetCommMask(hf_com, EV_RXCHAR);
	SetCommState(hf_com,&dcb);
	SetCommTimeouts(hf_com, &timeouts);
	EscapeCommFunction(hf_com, SETDTR);
	return 0;
}
static int matlabdcb(int nrhs, const mxArray*prhs[]){
	for(int j = 1;j < 6;j++) {
		if(!mxIsDouble(prhs[j])) {
			errorfunc(__LINE__);
			return -1;
		}
	}
	dcb.BaudRate = (int)mxGetScalar(prhs[1]);
	mexPrintf("BaudRate=%d,",dcb.BaudRate);
	dcb.ByteSize = (int)mxGetScalar(prhs[2]);
	mexPrintf("ByteSize=%d,",dcb.ByteSize);
	if((dcb.ByteSize < 7) || (dcb.ByteSize > 8)) {
		errorfunc(__LINE__);
		return -1;
	}
	dcb.Parity = (int)mxGetScalar(prhs[3]);
	if((dcb.Parity < 0) || (dcb.Parity >4)) {
		errorfunc(__LINE__);
		return -1;
	}
	switch(dcb.Parity) {
		case 0:mexPrintf("Parity=0(none),");break;
		case 1:mexPrintf("Parity=1(odd),");break;
		case 2:mexPrintf("Parity=2(even),");break;
		case 3:mexPrintf("Parity=3(mark),");break;
		case 4:mexPrintf("Parity=4(space),");break;
		default:mexPrintf("Parity=?,");
	}
	double tmpbit = 2.0*mxGetScalar(prhs[4]);
	if(tmpbit < 3.0) dcb.StopBits = 0;
	else if(tmpbit < 4.0) dcb.StopBits = 1;
	else dcb.StopBits = 2;
	if((dcb.StopBits < 0) || (dcb.StopBits >2)) {
		errorfunc(__LINE__);
		return -1;
	}
	switch(dcb.StopBits) {
		case 0:mexPrintf("StopBits=1(1bit),");break;
		case 1:mexPrintf("StopBits=1.5(1.5bit),");break;
		case 2:mexPrintf("StopBits=2(2bit),");break;
		default:mexPrintf("StopBits=?,");
	}
	flag = (int)mxGetScalar(prhs[5]);
// 0-> none, 1->Xon/Xoff, 2->hardware
	dcb.fDtrControl = DTR_CONTROL_ENABLE;
	dcb.fRtsControl = RTS_CONTROL_ENABLE;

#define CommXonLim 2048
#define CommXoffLim 2048
#define XON 0x11
#define XOFF 0x13
	
	switch(flag) {
		case 0x01:
			dcb.fOutX = TRUE;
			dcb.fInX = TRUE;
			dcb.XonLim = CommXonLim;
			dcb.XoffLim = CommXoffLim;
			dcb.XonChar = XON;
			dcb.XoffChar = XOFF;
			mexPrintf("Xon/Xoff=1,");
			break;
		case 0x02:
			dcb.fOutxCtsFlow = TRUE;
			dcb.fRtsControl = RTS_CONTROL_HANDSHAKE;
			mexPrintf("Hardware=2,");
			break;
		case 0x03:
			dcb.fDtrControl = DTR_CONTROL_DISABLE;
			dcb.fRtsControl = RTS_CONTROL_DISABLE;
			mexPrintf("DISABLE,");
			break;
		default: 
			mexPrintf("None=0,");
	}
	mexPrintf("Sleep=%d(mS)]\n",sleepinterval);
	return 0;
}
static int initCOM(char* Port) {
	if(initialized == 0) {
//		if(NULL ==(rdata=(unsigned char*)mxCalloc(BUFSIZE,sizeof(unsigned char)))) {
//			mexPrintf("Allocation Failed %s line %d %d\n",__FILE__,__LINE__,__DATE__);
//		}
//		mexMakeMemoryPersistent(rdata);
//		hf_com = CreateFile(Port, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED, NULL);
		hf_com = CreateFile(Port, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_FLAG_WRITE_THROUGH |FILE_FLAG_NO_BUFFERING | FILE_FLAG_OVERLAPPED, NULL);
		if(hf_com == INVALID_HANDLE_VALUE){
			CHAR ErrorBuf[256];
			wsprintf(ErrorBuf,"CreateFile Failed GetLastError = %d",GetLastError());
			mexPrintf("%s %s line %d %s\n",ErrorBuf,__FILE__,__LINE__,__DATE__);
			return -1;
		}
		DWORD DErr;
		ClearCommError(hf_com,&DErr,NULL);
		SetupComm(hf_com, BUFSIZE, BUFSIZE);
		PurgeComm(hf_com, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR ) ;
		SetCommTimeouts(hf_com, &timeouts);
		SetCommMask(hf_com,0);
		SetCommMask(hf_com, EV_RXCHAR);
		SetCommState(hf_com,&dcb);
		EscapeCommFunction(hf_com, SETDTR);
		switch(flag) {
			case 0x04:
				EscapeCommFunction(hf_com, CLRRTS);
				EscapeCommFunction(hf_com, CLRDTR);
				mexPrintf("Set[RTS=0,DTR=0]\n");
				break;
			case 0x05:
				EscapeCommFunction(hf_com, CLRRTS);
				EscapeCommFunction(hf_com, SETDTR);
				mexPrintf("Set[RTS=0,DTR=1]\n");
				break;
			case 0x06:
				EscapeCommFunction(hf_com, SETRTS);
				EscapeCommFunction(hf_com, CLRDTR);
				mexPrintf("Set[RTS=1,DTR=0]\n");
				break;
			case 0x07:
				EscapeCommFunction(hf_com, SETRTS);
				EscapeCommFunction(hf_com, SETDTR);
				mexPrintf("Set[RTS=1,DTR=1]\n");
				break;
		}
	}
	return 0;
}
volatile int terminate = 1;
unsigned char data[BUFSIZE + 1];
unsigned long len;
static unsigned long __stdcall readCOM(void* param) {
	int errornum;
	long length;
	DWORD dwErr;
	COMSTAT comstat;
	memset(&_olRead,0,sizeof(OVERLAPPED));
	_olRead.hEvent = CreateEvent(NULL,TRUE,FALSE,NULL);
	DWORD dwEvtMask = 0;
	SetCommMask(hf_com,EV_RXCHAR);
	rlen = 0;
	while(terminate) {
		dwEvtMask = 0;
		WaitCommEvent(hf_com, &dwEvtMask, &_olRead);
		ClearCommError(hf_com,&dwErr,&comstat);
		length = comstat.cbInQue;
		if(length == 0) {
			ResetEvent(_olRead.hEvent);
			Sleep(sleepinterval);
			continue;
		}
		if (length > BUFSIZE) length = BUFSIZE -1;
		if((dwEvtMask & EV_RXCHAR) || length) {
			EnterCriticalSection(&Cs);
			if(!ReadFile(hf_com,&data,length,&len,&_olRead)) {
				errornum = GetLastError();
				if (errornum == ERROR_IO_PENDING) {
					GetOverlappedResult(hf_com,&_olRead,(LPDWORD)&len,TRUE);
				} else if(errornum) mexPrintf("[%s]%d lasterror=%d\n",__FILE__,__LINE__,errornum);
			}
			for(int i = 0;i <(int) len;i++,rlen++) rdata[rlen & BUFSIZE] = data[i & BUFSIZE];
			LeaveCriticalSection(&Cs);
		}
		ResetEvent(_olRead.hEvent);
		Sleep(sleepinterval);
	}
	CloseHandle(_olRead.hEvent);
	return 0;
}
static void closeCOM(void) {
	if(initialized == 1) {
		initialized = 0;
		DWORD dwExitCode;
		terminate = 0;
		flag = 0;
		do {
			Sleep(500);
			GetExitCodeThread(hThrd,&dwExitCode);
		} while(dwExitCode == STILL_ACTIVE);
//		mxFree(rdata);
		EscapeCommFunction(hf_com, CLRDTR );
		SetCommMask(hf_com,0);
		CloseHandle(_olWrite.hEvent);
		PurgeComm(hf_com, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR ) ;
		CloseHandle(hf_com);
		DeleteCriticalSection(&Cs);
		showusage();
	}
	return;
}
/*  Support Usage : 
mode 0: init                    rs232('COM1',BaudRate,ByteSize,Parity,StopBits,Flwctl,Sleep)
mode 1: getc               		uint8_data = rs232([]);
mode 2: putc                    status = rs232(uint8_data);
*/
void mexFunction( int nlhs, mxArray *plhs[],int nrhs, const mxArray*prhs[] ){
	char Port[256];
	double *cts,*dsr,*ring,*cd;
	int buflen,status;
	int length;
	COMSTAT comstat;
	DWORD dwErr = 0;
	if(!initialized) {
		if(nrhs == 6);
		else {
			if(nrhs == 7) {
				sleepinterval = (int)mxGetScalar(prhs[6]);
			}
			else {
				errorfunc(__LINE__);
				return;
			}
		}
		buflen=(mxGetM(prhs[0])*mxGetN(prhs[0]))+1;
		status = 1;
		switch(buflen) {
			case 5:
				status = mxGetString(prhs[0],Port,buflen);
				break;
			case 6:
				Port[0]=92;Port[1]=92;
				Port[2]=46;Port[3]=92;
				status = mxGetString(prhs[0],&Port[4],buflen);
				break;
			default:
				errorfunc(__LINE__);
				return;
		}
		if(status != 0) {errorfunc(__LINE__);return;}
		mexPrintf("[Port=%s,",Port);
		if(initdcb(Port) != 0){errorfunc(__LINE__);return;}
		if(matlabdcb(nrhs,prhs) != 0){errorfunc(__LINE__);return;}
		if(initCOM(Port) != 0){errorfunc(__LINE__);return;}
		InitializeCriticalSection(&Cs);
		mexAtExit(closeCOM);
		initialized = 1;
//		hThrd = (HANDLE)CreateThread(NULL,0,readCOM,0,0,&dwThrdID);
		hThrd = (HANDLE)_beginthreadex(NULL,0,(unsigned int (__stdcall *)(void *))readCOM,0,0,&dwThrdID);
//		SetThreadPriority(&dwThrdID,-15);
		DWORD dwErr = 0;
		ClearCommError(hf_com,&dwErr,&comstat);
		memset(&_olWrite,0,sizeof(OVERLAPPED));
		if(NULL == (_olWrite.hEvent = CreateEvent(NULL,TRUE,FALSE,NULL))) {
			mexPrintf("Event Err %s line %d %d\n",__FILE__,__LINE__,__DATE__);
			return;
		}
		return;
	}
	if((nlhs == 0)&&(nrhs == 0)) {errorfunc(__LINE__);return;}
	if(nlhs == 4) {
		DWORD x00;
		GetCommModemStatus(hf_com,&x00);
		plhs[0] = mxCreateDoubleMatrix(1,1,mxREAL);
		plhs[1] = mxCreateDoubleMatrix(1,1,mxREAL);
		plhs[2] = mxCreateDoubleMatrix(1,1,mxREAL);
		plhs[3] = mxCreateDoubleMatrix(1,1,mxREAL);
		cts = mxGetPr(plhs[0]);
		dsr = mxGetPr(plhs[1]);
		ring = mxGetPr(plhs[2]);
		cd = mxGetPr(plhs[3]);
		cts[0]=0.;if(x00 & MS_CTS_ON) cts[0]=1.;
		dsr[0]=0.;if(x00 & MS_DSR_ON) dsr[0]=1.;
		ring[0]=0.;if(x00 & MS_RING_ON) ring[0]=1.;
		cd[0]=0.;if(x00 & MS_RLSD_ON) cd[0]=1.;
		return;
	}
// READ MODE
	if(mxGetN(prhs[0]) * mxGetM(prhs[0]) == 0) {
			EnterCriticalSection(&Cs);
			if(rlen > BUFSIZE) plhs[0] = mxCreateDoubleMatrix(1,BUFSIZE,mxREAL);
			else plhs[0] = mxCreateDoubleMatrix(1,rlen,mxREAL);
			double *out_data = mxGetPr(plhs[0]);
			int i;
			if(rlen <= BUFSIZE) for(i = 0;i < rlen;i++) out_data[i] = (double)rdata[i & BUFSIZE];
			else for(i = 0;i < BUFSIZE;i++) out_data[i] = (double)rdata[(rlen + i + 1) & BUFSIZE];
			rlen = 0;
			LeaveCriticalSection(&Cs);
			return;
	} else {
		if(!mxIsUint8(prhs[0])) {
			mexPrintf("This program supports only uint8 array data\n");
			return;
		}
// WRITE MODE
		int errornum;
		unsigned long wlen = (unsigned long)mxGetM(prhs[0])*mxGetN(prhs[0]);
		unsigned char *wdata = (unsigned char*)mxGetPr(prhs[0]);
		ResetEvent(_olWrite.hEvent);
		int ret = WriteFile(hf_com,wdata,wlen,&wlen,&_olWrite);
//		FlushFileBuffers(hf_com);
		if (!ret) {
			errornum = GetLastError();
			if (errornum == ERROR_IO_PENDING) {
				GetOverlappedResult(hf_com,&_olWrite,(LPDWORD)&wlen,TRUE);
			} else if(errornum) mexPrintf("[%s]%d lasterror=%d\n",__FILE__,__LINE__,errornum);
		} else mexPrintf("[%s]%d lasterror=%d\n",__FILE__,__LINE__,errornum);
		FlushFileBuffers(hf_com);
		return;
	}
}
